/**
 * @license
 * Copyright (c) 2016 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at http://polymer.github.io/PATENTS.txt
 */
"use strict";
const hydrolysis_1 = require('hydrolysis');
const path_1 = require('path');
const osPath = require('path');
const stream_1 = require('stream');
const url_1 = require('url');
const logging = require('plylog');
const dom5_1 = require('dom5');
const minimatchAll = require('minimatch-all');
const logger = logging.getLogger('cli.build.analyzer');
class StreamAnalyzer extends stream_1.Transform {
    constructor(root, entrypoint, shell, fragments) {
        super({ objectMode: true });
        this.files = new Map();
        this.root = root;
        this.entrypoint = entrypoint;
        this.shell = shell;
        this.fragments = fragments;
        this.allFragments = [];
        // It's important that shell is first for document-ordering of imports
        if (shell) {
            this.allFragments.push(shell);
        }
        if (entrypoint && !shell && fragments.length === 0) {
            this.allFragments.push(entrypoint);
        }
        if (fragments) {
            this.allFragments = this.allFragments.concat(fragments);
        }
        this.resolver = new StreamResolver(this);
        this.loader = new hydrolysis_1.Loader();
        this.loader.addResolver(this.resolver);
        this.analyzer = new hydrolysis_1.Analyzer(false, this.loader);
        this.analyze = new Promise((resolve, reject) => {
            this._analyzeResolve = resolve;
        });
    }
    _transform(file, encoding, callback) {
        this.addFile(file);
        // If this is the entrypoint, hold on to the file, so that it's fully
        // analyzed by the time down-stream transforms see it.
        if (this.isFragment(file)) {
            callback(null, null);
        }
        else {
            callback(null, file);
        }
    }
    _flush(done) {
        this._getDepsToEntrypointIndex().then((depsIndex) => {
            // push held back files
            for (let fragment of this.allFragments) {
                let url = this.urlFromPath(fragment);
                let file = this.getUrl(url);
                if (file == null) {
                    done(new Error(`no file found for fragment ${fragment}`));
                }
                this.push(file);
            }
            this._analyzeResolve(depsIndex);
            done();
        });
    }
    /**
     * A side-channel to add files to the resolver that did not come throgh the
     * stream transformation. This is for generated files, like
     * shared-bundle.html. This should probably be refactored so that the files
     * can be injected into the stream.
     */
    addFile(file) {
        // Badly-behaved upstream transformers (looking at you gulp-html-minifier)
        // may use posix path separators on Windows.
        let filepath = osPath.normalize(file.path);
        // Store only root-relative paths, in URL/posix format
        this.files.set(this.urlFromPath(filepath), file);
    }
    getFile(filepath) {
        return this.getUrl(this.urlFromPath(filepath));
    }
    getUrl(url) {
        if (url.startsWith('/')) {
            url = url.substring(1);
        }
        let file = this.files.get(url);
        if (!file) {
            logger.debug(`no file for ${url} :(`);
        }
        return file;
    }
    isFragment(file) {
        return this.allFragments.indexOf(file.path) !== -1;
    }
    urlFromPath(filepath) {
        if (!filepath.startsWith(this.root)) {
            throw new Error(`file path is not in root: ${filepath} (${this.root})`);
        }
        // convert filesystem path to URL
        let relativePath = osPath.relative(this.root, filepath);
        let url = path_1.posix.join.apply(null, relativePath.split(osPath.sep));
        return url;
    }
    _getDepsToEntrypointIndex() {
        let depsPromises = this.allFragments.map((f) => this._getDependencies(this.urlFromPath(f)));
        return Promise.all(depsPromises).then((value) => {
            // tsc was giving a spurious error with `allDeps` as the parameter
            let allDeps = value;
            // An index of dependency -> fragments that depend on it
            let depsToFragments = new Map();
            // An index of fragments -> dependencies
            let fragmentToDeps = new Map();
            let fragmentToFullDeps = new Map();
            console.assert(this.allFragments.length === allDeps.length);
            for (let i = 0; i < allDeps.length; i++) {
                let fragment = this.allFragments[i];
                let deps = allDeps[i];
                console.assert(deps != null, `deps is null for ${fragment}`);
                fragmentToDeps.set(fragment, deps.imports);
                fragmentToFullDeps.set(fragment, deps);
                for (let dep of deps.imports) {
                    let entrypointList;
                    if (!depsToFragments.has(dep)) {
                        entrypointList = [];
                        depsToFragments.set(dep, entrypointList);
                    }
                    else {
                        entrypointList = depsToFragments.get(dep);
                    }
                    entrypointList.push(fragment);
                }
            }
            return {
                depsToFragments: depsToFragments,
                fragmentToDeps: fragmentToDeps,
                fragmentToFullDeps: fragmentToFullDeps,
            };
        });
    }
    /**
     * Attempts to retreive document-order transitive dependencies for `url`.
     */
    _getDependencies(url) {
        let dir = path_1.posix.dirname(url);
        return this.analyzer.metadataTree(url)
            .then((tree) => this._getDependenciesFromDescriptor(tree, dir));
    }
    _getDependenciesFromDescriptor(descriptor, dir) {
        let allHtmlDeps = [];
        let allScriptDeps = new Set();
        let allStyleDeps = new Set();
        let deps = this._collectScriptsAndStyles(descriptor);
        deps.scripts.forEach((s) => allScriptDeps.add(path_1.posix.resolve(dir, s)));
        deps.styles.forEach((s) => allStyleDeps.add(path_1.posix.resolve(dir, s)));
        if (descriptor.imports) {
            let queue = descriptor.imports.slice();
            while (queue.length > 0) {
                let next = queue.shift();
                if (!next.href) {
                    continue;
                }
                allHtmlDeps.push(next.href);
                let childDeps = this._getDependenciesFromDescriptor(next, path_1.posix.dirname(next.href));
                allHtmlDeps = allHtmlDeps.concat(childDeps.imports);
                childDeps.scripts.forEach((s) => allScriptDeps.add(s));
                childDeps.styles.forEach((s) => allStyleDeps.add(s));
            }
        }
        return {
            scripts: Array.from(allScriptDeps),
            styles: Array.from(allStyleDeps),
            imports: allHtmlDeps,
        };
    }
    _collectScriptsAndStyles(tree) {
        let scripts = [];
        let styles = [];
        tree.html.script.forEach((script) => {
            if (script['__hydrolysisInlined']) {
                scripts.push(script['__hydrolysisInlined']);
            }
        });
        tree.html.style.forEach((style) => {
            let href = dom5_1.getAttribute(style, 'href');
            if (href) {
                styles.push(href);
            }
        });
        return {
            scripts: scripts,
            styles: styles
        };
    }
}
exports.StreamAnalyzer = StreamAnalyzer;
class StreamResolver {
    constructor(analyzer) {
        this.analyzer = analyzer;
    }
    accept(url, deferred) {
        let urlObject = url_1.parse(url);
        if (urlObject.hostname || !urlObject.pathname) {
            return false;
        }
        let urlPath = decodeURIComponent(urlObject.pathname);
        let file = this.analyzer.getUrl(urlPath);
        if (file) {
            deferred.resolve(file.contents.toString());
        }
        else {
            logger.debug(`No file found for ${urlPath}`);
            // If you're template to do the next line, Loader does that for us, so
            // don't double reject!
            // deferred.reject(new Error(`No file found for ${urlPath}`));
            return false;
        }
        return true;
    }
}